#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h> 

#define N 80
#define PORTA "50000"

/* Gestore del segnale SIGCHLD */
void handler(int s)
{
	int status;
	while(waitpid(-1, &status, WNOHANG) > 0);
}


int main(int argc, char** argv)
{
	struct addrinfo hints, *res;
	int err, sd, ns, pid, on;
	char option[N];

	/* Controllo argomenti */
	if(argc > 1){
		printf("Uso: ./server\n");
		exit(1);
	}

	memset(&hints, 0, sizeof(hints));
	hints.ai_family   = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags    = AI_PASSIVE;

	if ((err = getaddrinfo(NULL, PORTA, &hints, &res)) != 0) {
		fprintf(stderr, "Errore setup indirizzo bind: %s\n", gai_strerror(err));
		exit(2);
	}

	if ((sd = socket(res->ai_family, res->ai_socktype, res->ai_protocol)) < 0){
		perror("Errore in socket");
		exit(3);
	}

	on = 1;
	if(setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on))<0){
		perror("setsockopt");
		exit(4);
	}

	if (bind(sd, res->ai_addr, res->ai_addrlen) < 0) {
		perror("Errore in bind");
		exit(5);
	}

	freeaddrinfo(res);
	
	if(listen(sd, SOMAXCONN)<0){
		perror("listen");
		exit(6);
	}
	
	/* The C library function void (*signal(int sig, void (*func)(int)))(int) 
	sets a function to handle signal, i.e., a signal handler with signal number sig. */
	signal(SIGCHLD, handler);

	/* Attendo i client... */
	for(;;){
		int nread;
		printf("Server in ascolto...\n");

		if((ns = accept(sd, NULL, NULL)) < 0) {
			/* Ignoro gli errori di tipo interruzione da segnale. */
			/* The <errno.h> header file defines the integer variable errno, which
			is set by system calls and some library functions in the event of an
			error to indicate what went wrong.  Its value is significant only
			when the return value of the call indicated an error (i.e., -1 from
			most system calls; -1 or NULL from most library functions); a
			function that succeeds is allowed to change errno. */
			/* EINTR Interrupted function call */
			if (errno == EINTR)
				continue;
			/* Gestisco tutte le altre tipologie di errore. */
			perror("accept");
			exit(7);
		}
	
		/* Generazione di un figlio */
		if((pid = fork())<0){
			perror("fork");
			exit(8);
		} else if(pid == 0){
			/* figlio */
			close(sd);

			/* Leggo option  */
			memset(option, 0, sizeof(option));
			if((nread=read(ns, option, N-1))<0){
				perror("read option");
				exit(9);
			}

			printf("nread %d\n",nread);
            printf("Eseguo richiesta con argomenti: *%s*\n",option);
            fflush(stdout);
			/* redirezione stdout e stderr */
			/* 0	Standard input 
			   1	Standard output 
			   2	Standard error */
			close(1);
			close(2);
			/*
			The dup() system call creates a copy of the file descriptor oldfd,
			using the lowest-numbered unused file descriptor for the new descriptor. */
			dup(ns);
			dup(ns);
			close(ns);
			
			/* Eseguo il ps sul server con i parametri corretti */
			/* An array of pointers to null-terminated strings that represent 
			the argument list available to the new program. The first argument, 
			by convention, should point to the filename associated with the file being executed. 
			The array of pointers must be terminated by a NULL pointer. */
			if (strlen(option)==0){ 
				execlp("ps", "ps", (char*)0);
				perror("exec ps 1");
				exit(10);
			} else {
				execlp("ps", "ps", option, (char*)0);
				perror("exec ps 2");
				exit(11);
			}
		} else {
			/* padre */
			close(ns);
		}
	}
	return 0;
}
