#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>

#define N 80

/* Gestore del segnale SIGCHLD */
void handler(int s)
{
	int status;
	wait(&status);
}


int main(int argc, char** argv)
{
	struct addrinfo hints, *res;
	int err, sd, ns, pid, on, file;
	char nomeFile[N], stringa[N];
    char *ack="ack\n";

	/* Controllo argomenti */
	if(argc < 2){
		printf("Uso: ./server <porta> \n");
		exit(1);
	}

	signal(SIGCHLD, handler);
	
	memset(&hints, 0, sizeof(hints));
	hints.ai_family   = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags    = AI_PASSIVE;

	if ((err = getaddrinfo(NULL, argv[1], &hints, &res)) != 0) {
		fprintf(stderr, "Errore setup indirizzo bind: %s\n", gai_strerror(err));
		exit(2);
	}

	if ((sd = socket(res->ai_family, res->ai_socktype, res->ai_protocol)) < 0){
		perror("Errore in socket");
		exit(3);
	}

	on = 1;
	if(setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on))<0){
		perror("setsockopt");
		exit(4);
	}

	if (bind(sd, res->ai_addr, res->ai_addrlen) < 0) {
		perror("Errore in bind");
		exit(5);
	}

	freeaddrinfo(res);
	
	if(listen(sd, SOMAXCONN)<0){
		perror("listen");
		exit(6);
	}

	signal(SIGCHLD, handler);

	/* Attendo i client... */
	for(;;){
		printf("Server in ascolto...\n");

		if((ns = accept(sd, NULL, NULL)) < 0) {
			/* Ignoro gli errori di tipo interruzione da segnale. */
			if (errno == EINTR)
				continue;
			/* Gestisco tutte le altre tipologie di errore. */
			perror("accept");
			exit(7);
		}
	
		/* Generazione di un figlio */
		if((pid = fork())<0){
			perror("fork");
			exit(8);
		} else if(pid == 0){
			/* figlio */
			close(sd);

			memset(nomeFile, 0, sizeof(nomeFile));
			if(read(ns, nomeFile, N-1)<0){
				perror("read nomeFile");
				exit(9);
			}
			printf("read nomeFile %s\n", nomeFile);

			if((file=(open(nomeFile, O_RDONLY)))==-1){
				//non trovo il file, chiudo la socket e termino il programma
				close(ns); // cosa succede lato client alla chiusura della socket?
				exit(9);
			}
			else{
				close(file);
			}
			
			if(write(ns, ack, strlen(ack))<0){
				perror("scrivo ack");
				exit(12);
			}
			
            memset(stringa, 0, sizeof(stringa));
			if(read(ns, stringa, N-1)<0){
				perror("read stringa");
				exit(13);
			}
							
			/* ridireziono stdout e stderr */
			close(1);
			close(2);
			dup(ns);
			dup(ns);
			close(ns);
			
			/* Eseguo la grep sul server con i parametri corretti */
			execlp("grep", "grep", stringa, nomeFile, (char*)0);
			perror("exec grep");
			exit(10);
			
		} else {
			/* padre */
			close(ns);
		}
	}
	return 0;
}
